/* ======================================================================== *\
!
! *
! * This file is part of MARS, the MAGIC Analysis and Reconstruction
! * Software. It is distributed to you in the hope that it can be a useful
! * and timesaving tool in analysing Data of imaging Cerenkov telescopes.
! * It is distributed WITHOUT ANY WARRANTY.
! *
! * Permission to use, copy, modify and distribute this software and its
! * documentation for any purpose is hereby granted without fee,
! * provided that the above copyright notice appear in all copies and
! * that both that copyright notice and this permission notice appear
! * in supporting documentation. It is provided "as is" without express
! * or implied warranty.
! *
!
!
!   Author(s): Thomas Bretz, 10/2001 <mailto:tbretz@astro.uni-wuerzburg.de>
!
!   Copyright: MAGIC Software Development, 2000-2003
!
!
\* ======================================================================== */
#include "MEventDisplay.h"

//
// C-lib
//
#include <stdlib.h>              // atoi

//
// root
//
#include <TFile.h>               // TFile
#include <TTree.h>               // TTree
#include <TList.h>               // TList::Add
#include <TStyle.h>              // gStyle->SetOptStat
#include <TCanvas.h>             // TCanvas::cd
#include <TSystem.h>             // TSystem::BaseName

//
// root GUI
//
#include <TGLabel.h>             // TGLabel
#include <TGButton.h>            // TGPictureButton
#include <TG3DLine.h>            // TGHorizontal3DLine
#include <TGTextEntry.h>         // TGTextEntry
#include <TGButtonGroup.h>       // TGVButtonGroup
#include <TRootEmbeddedCanvas.h> // TRootEmbeddedCanvas

//
// General
//
#include "MGList.h"              // MGList

#include "MLog.h"
#include "MLogManip.h"

#include "MParList.h"            // MParList::AddToList
#include "MEvtLoop.h"            // MEvtLoop::GetParList
#include "MTaskList.h"           // MTaskList::AddToList

//
// Tasks
//
#include "MReadMarsFile.h"        // MReadMarsFile
#include "MGeomApply.h"           // MGeomApply
#include "MFDataMember.h"         // MFDataMember
#include "MMcPedestalCopy.h"      // MMcPedestalCopy
#include "MMcPedestalNSBAdd.h"    // MMcPedestalNSBAdd

#include "MSignalCalc.h"          // MPedestalCalc
#include "MTaskEnv.h"             // MTaskEnv
#include "MExtractTimeAndChargeDigitalFilter.h"
#include "MImgCleanStd.h"         // MImgCleanStd
#include "MHillasCalc.h"          // MHillasCalc
#include "MCalibrationRelTimeCalc.h" // MHillasSrcCalc
#include "MPrint.h"               // 
#include "MBadPixelsCalc.h"       // MBadPixelsCalc
#include "MBadPixelsTreat.h"      // MBadPixelsTreat
#include "MFillH.h"               // MFillH
#include "MExtractSignal.h"       // MExtractsignal
#include "MMcCalibrationUpdate.h" // MMcCalibrationUpdate
#include "MCalibrateData.h"       // MCalibrateData
#include "MMuonSearchParCalc.h"   // MMuonSearchParCalc
#include "MMuonCalibParCalc.h"    // MMuonCalibParCalc
//#include "MMcTriggerLvl2Calc.h"   // MMcTriggerLvl2Calc

//
// Container
//
#include "MHillas.h"               // MHillas::Print(const MGeomCam&)
#include "MHillasExt.h"            // MHillasExt::Print(const MGeomCam&)
#include "MHillasSrc.h"            // MHillasSrc::Print(const MGeomCam&)
#include "MImagePar.h"             // MImagePar::Print(const MGeomCam&)
#include "MNewImagePar.h"          // MNewImagePar::Print(const MGeomCam&)
#include "MHEvent.h"               // MHEvent
#include "MHCamera.h"              // MHCamera
#include "MRawEvtData.h"           // MRawEvtData
#include "MArrivalTimeCam.h"       // MArrivalTimeCam
#include "MBadPixelsCam.h"         // MBadPixelsCam
#include "MPedPhotCam.h"           // MPedPhotCam
#include "MCalibrationChargeCam.h" // MCalibrationChargeCam
#include "MMuonSearchPar.h"        // MMuonCalibPar
//#include "MMcTriggerLvl2.h"        // MMcTriggerLvl2

using namespace std;

ClassImp(MEventDisplay);

// --------------------------------------------------------------------------
//
//  Constructor.
//
MEventDisplay::MEventDisplay(const char *fname) : MStatusDisplay(), fEvtLoop(0)
{
    //
    // Setup Task list for hillas calculation
    //
    SetupTaskList("Events", fname);

    //
    // Add MEventDisplay GUI elements to the display
    //
    AddUserFrame(fname);

    //
    // Show new part of the window, resize to correct aspect ratio
    //
    // FIXME: This should be done by MStatusDisplay automatically
    Resize(GetWidth(), GetHeight() + fUserFrame->GetDefaultHeight());
    SetWindowName("Event Display");
    MapSubwindows();

    //
    // Readin first event and display it
    //
    if (fEvtLoop)
        ReadFirstEvent();

    SetNoContextMenu(kFALSE);
}

// --------------------------------------------------------------------------
//
//  Destructor: PostProcess eventloop, delete eventloop with all containers
//
MEventDisplay::~MEventDisplay()
{
    if (fEvtLoop)
    {
        fEvtLoop->PostProcess();
        delete fEvtLoop;
    }
}

Int_t MEventDisplay::GetFileType(const char *tree, const char *fname) const
{
    TFile f(fname, "READ");
    TTree *t = (TTree*)f.Get(tree);
    if (!t)
        return -1;

    if (t->FindBranch("MSignalCam."))
        return 1;

    if (t->FindBranch("MRawEvtData."))
        return 2;

    return -2;
}

// --------------------------------------------------------------------------
//
//  Setup Task and parameter list for hillas calculation,
//  preprocess tasks and read in first event (process)
//
void MEventDisplay::SetupTaskList(const char *tname, const char *fname)
{
//    MCalibrationChargeCam *ccam=0;
//    MPedPhotCam           *pcam=0;

    MBadPixelsCam *badpix = new MBadPixelsCam;

    const Int_t type = GetFileType(tname, fname);
    switch (type)
    {
    case  1: gLog << all << "Calibrated Data File detected..." << endl; break;
    case  2: gLog << all << "Raw Data File detected..." << endl; break;
    case -2: gLog << err << "File type unknown... abort." << endl; return;
    case -1: gLog << err << "Tree " << tname << " not found... abort." << endl; return;
    }

    //
    // Setup an empty job, with a reader task only.
    // All tasks and parameter containers are deleted automatically
    // (via SetOwner())
    //
    MTaskList *tlist = new MTaskList;
    tlist->SetOwner();

    MReadMarsFile *read = new MReadMarsFile(tname, fname);
    read->DisableAutoScheme();
    tlist->AddToList(read);

    MGeomApply *apl = new MGeomApply;
    tlist->AddToList(apl);

    MParList *plist = new MParList;
    plist->SetOwner();
    plist->AddToList(tlist);
    plist->AddToList(badpix);

    //MArrivalTime *atime = new MArrivalTime;
    //plist->AddToList(atime);

    MHEvent *evt01 = new MHEvent(MHEvent::kEvtSignalDensity);
    MHEvent *evt02 = new MHEvent(MHEvent::kEvtSignalDensity);
    MHEvent *evt03 = new MHEvent(type==1?MHEvent::kEvtPedPhot:MHEvent::kEvtPedestal);
    MHEvent *evt04 = new MHEvent(type==1?MHEvent::kEvtPedPhotRMS:MHEvent::kEvtPedestalRMS);
    MHEvent *evt06a= new MHEvent(MHEvent::kEvtCleaningData);
    MHEvent *evt06b= new MHEvent(MHEvent::kEvtCleaningLevels);
    MHEvent *evt07 = new MHEvent(MHEvent::kEvtIdxMax);
    MHEvent *evt08 = new MHEvent(MHEvent::kEvtArrTime);
    //MHEvent *evt09 = new MHEvent(MHEvent::kEvtTrigPix);
    MHEvent *evt10 = new MHEvent(MHEvent::kEvtIslandIndex);

    evt01->SetName("Signal");
    evt02->SetName("Cleaned");
    evt03->SetName("Pedestal");
    evt04->SetName("PedRMS");
    evt06a->SetName("CleanData");
    evt06b->SetName("CleanLevels");
    evt07->SetName("Max Slice Idx");
    evt08->SetName("Arrival Time");
    //evt09->SetName("Trigger");
    evt10->SetName("Islands");

    // This makes sure, that the containers are deleted...
    plist->AddToList(evt01);
    plist->AddToList(evt02);
    plist->AddToList(evt03);
    plist->AddToList(evt04);
    plist->AddToList(evt06a);
    plist->AddToList(evt06b);
    plist->AddToList(evt07);
    plist->AddToList(evt08);
    //plist->AddToList(evt09);
    plist->AddToList(evt10);

    MExtractTimeAndChargeDigitalFilter *digf = new MExtractTimeAndChargeDigitalFilter;
    digf->SetNameWeightsFile("msignal/cosmics_weights46.dat");

    MTaskEnv *taskenv1=new MTaskEnv("ExtractSignal");
    taskenv1->SetDefault(digf);
    taskenv1->SetOwner();

    MSignalCalc     *nanal   = new MSignalCalc;
    MFillH          *fill01  = new MFillH(evt01, "MSignalCam", "MFillH01");
    MImgCleanStd    *clean   = new MImgCleanStd;
    MFillH          *fill02  = new MFillH(evt02, "MSignalCam", "MFillH02");
    MFillH          *fill03  = new MFillH(evt03, type==1?"MPedPhotFundamental":"MPedestalCam",       "MFillH03");
    MFillH          *fill04  = new MFillH(evt04, type==1?"MPedPhotFromExtractorRndm":"MPedestalCam", "MFillH04");
    MFillH          *fill06a = new MFillH(evt06a, "MCameraData", "MFillH06a");
    MFillH          *fill06b = new MFillH(evt06b, "MCameraData", "MFillH06b");
    MHillasCalc     *hcalc   = new MHillasCalc;
    //MMcTriggerLvl2Calc *trcal  = new MMcTriggerLvl2Calc;
    //MFillH             *fill09 = new MFillH(evt09, "MMcTriggerLvl2", "MFillH09");
    MFillH          *fill10  = new MFillH(evt10, "MSignalCam",    "MFillH10");

    MBadPixelsCalc  *bcalc   = new MBadPixelsCalc;
    MBadPixelsTreat *btreat  = new MBadPixelsTreat;
    btreat->SetProcessTimes(kFALSE);
    if (type==1)
    {
        bcalc->SetNamePedPhotCam("MPedPhotFromExtractor");
        btreat->AddNamePedPhotCam("MPedPhotFromExtractorRndm");
        clean->SetNamePedPhotCam("MPedPhotFromExtractorRndm");
    }

    // If no pedestal or no calibration file is availble
    if (type==2)
    {
        MFilter *f1=new MFDataMember("MRawRunHeader.fRunType", '>', 255.5);
        f1->SetName("MFMonteCarlo");

        MMcPedestalCopy   *pcopy = new MMcPedestalCopy;
        MMcPedestalNSBAdd *pdnsb = new MMcPedestalNSBAdd;

        MMcCalibrationUpdate* mcupd = new MMcCalibrationUpdate;
        mcupd->SetOuterPixelsGainScaling(kFALSE);

        MCalibrateData* calib = new MCalibrateData;
        calib->SetCalibrationMode(MCalibrateData::kFlatCharge);
        calib->SetPedestalFlag(MCalibrateData::kEvent);

        //MCalibrationRelTimeCalc *tcalc = new MCalibrationRelTimeCalc;

        // MC
        mcupd->SetFilter(f1);
        //trcal->SetFilter(f1);

        // TaskList
        tlist->AddToList(f1);
        tlist->AddToList(pcopy);
        tlist->AddToList(pdnsb);

        tlist->AddToList(nanal);    // Calculated MPedPhotCam
        tlist->AddToList(taskenv1); // Calculates MExtractedSignalCam, MArrivalTimeCam
        tlist->AddToList(mcupd);
        tlist->AddToList(calib);    // MExtractedSignalCam --> MSignalCam

        tlist->AddToList(bcalc);    // Produce MBadPixelsCam
        tlist->AddToList(btreat);   // Treat MBadPixelsCam
    }

    tlist->AddToList(fill01);
    tlist->AddToList(clean);
    tlist->AddToList(fill02);
    tlist->AddToList(fill03);
    tlist->AddToList(fill04);
    tlist->AddToList(fill06a);
    tlist->AddToList(fill06b);
    tlist->AddToList(fill10);
    if (type==2)
    {
        MFillH *fill07 = new MFillH(evt07, "MRawEvtData",     "MFillH7");
        MFillH *fill08 = new MFillH(evt08, "MArrivalTimeCam", "MFillH8");
        tlist->AddToList(fill07);
        tlist->AddToList(fill08);

	//tlist->AddToList(trcal);
	//tlist->AddToList(fill09);
    }
    if (type==1)
    {
        MFillH *fill08 = new MFillH(evt08, "MSignalCam", "MFillH8");
        tlist->AddToList(fill08);
    }
    tlist->AddToList(hcalc);
    // --------------------------------------------------
    MMuonSearchParCalc *muscalc = new MMuonSearchParCalc;
    MMuonCalibParCalc  *mcalc   = new MMuonCalibParCalc;
    MFillH *fillmuon = new MFillH("MHSingleMuon", "", "FillMuon");
    fillmuon->SetNameTab("Muon");
    tlist->AddToList(muscalc);
    tlist->AddToList(fillmuon);
    tlist->AddToList(mcalc);
    // --------------------------------------------------

    //
    // Now distribute Display to all tasks
    //
    fEvtLoop = new MEvtLoop(gSystem->BaseName(fname));
    fEvtLoop->SetOwner();
    fEvtLoop->SetParList(plist);
    fEvtLoop->SetDisplay(this);
    fEvtLoop->ReadEnv("mars.rc");

}

// --------------------------------------------------------------------------
//
//  Add the top part of the frame: This is filename and treename display
//
void MEventDisplay::AddTopFramePart1(TGCompositeFrame *frame,
                                    const char *filename,
                                    const char *treename)
{
    //
    //  --- the top1 part of the window ---
    //
    TGHorizontalFrame *top1 = new TGHorizontalFrame(frame, 1, 1);
    fList->Add(top1);

    //
    // create gui elements
    //
    TGLabel *file = new TGLabel(top1, new TGString(Form("%s#%s", filename, treename)));
    fList->Add(file);

    //
    // layout and add gui elements in/to frame
    //
    TGLayoutHints *laystd = new TGLayoutHints(kLHintsCenterX, 5, 5);
    fList->Add(laystd);

    top1->AddFrame(file,  laystd);

    //
    //  --- the top1 part of the window ---
    //
    TGHorizontalFrame *top2 = new TGHorizontalFrame(frame, 1, 1);
    fList->Add(top2);

    //
    // layout and add frames
    //
    TGLayoutHints *laytop1 = new TGLayoutHints(kLHintsCenterX, 5, 5, 5);
    fList->Add(laytop1);
    frame->AddFrame(top1, laytop1);
    frame->AddFrame(top2, laytop1);
}

// --------------------------------------------------------------------------
//
//  Add the second part of the top frame: This are the event number controls
//
void MEventDisplay::AddTopFramePart2(TGCompositeFrame *frame)
{
    //
    // --- the top2 part of the window ---
    //
    TGHorizontalFrame *top2 = new TGHorizontalFrame(frame, 1, 1);
    fList->Add(top2);

    //
    // Create the gui elements
    //
    TGTextButton *prevevt = new TGTextButton(top2, " << ", kEvtPrev);
    prevevt->Associate(this);

    TGLabel *evtnr = new TGLabel(top2, new TGString("Event:"));

    TGTextEntry *entry=new TGTextEntry(top2, new TGTextBuffer(100), kEvtNumber);
    entry->Resize(60, entry->GetDefaultHeight());
    entry->Associate(this);

    fNumOfEvts = new TGLabel(top2, "of           .");

    TGTextButton *nextevt = new TGTextButton (top2, " >> ", kEvtNext);
    nextevt->Associate(this);

    //
    // Add gui elements to 'atotodel'
    //
    fList->Add(prevevt);
    fList->Add(evtnr);
    fList->Add(entry);
    fList->Add(fNumOfEvts);
    fList->Add(nextevt);

    //
    // add the gui elements to the frame
    //
    TGLayoutHints *laystd = new TGLayoutHints(kLHintsLeft|kLHintsCenterY, 5, 5);
    fList->Add(laystd);

    top2->AddFrame(prevevt,    laystd);
    top2->AddFrame(evtnr,      laystd);
    top2->AddFrame(entry,      laystd);
    top2->AddFrame(fNumOfEvts, laystd);
    top2->AddFrame(nextevt,    laystd);

    TGLayoutHints *laystd2 = new TGLayoutHints(kLHintsCenterX, 5, 5, 5, 5);
    fList->Add(laystd2);
    frame->AddFrame(top2, laystd2);

    //
    // Add trailing line...
    //
    TGHorizontal3DLine *line = new TGHorizontal3DLine(frame);
    TGLayoutHints *layline = new TGLayoutHints(kLHintsExpandX);
    fList->Add(line);
    fList->Add(layline);
    frame->AddFrame(line, layline);
}

// --------------------------------------------------------------------------
//
//  Add the user frame part of the display
//
void MEventDisplay::AddUserFrame(const char* filename)
{
    fUserFrame->ChangeOptions(kHorizontalFrame);

    TGCompositeFrame *vf1 = new TGVerticalFrame(fUserFrame, 1, 1);
    TGCompositeFrame *vf2 = new TGVerticalFrame(fUserFrame, 1, 1);

    AddTopFramePart1(vf1, filename, "Events");
    AddTopFramePart2(vf1);

    // create root embedded canvas and add it to the tab
    TRootEmbeddedCanvas *ec = new TRootEmbeddedCanvas("Slices", vf2, vf1->GetDefaultHeight()*3/2, vf1->GetDefaultHeight(), 0);
    vf2->AddFrame(ec);
    fList->Add(ec);

    // set background and border mode of the canvas
    fCanvas = ec->GetCanvas();
    fCanvas->SetBorderMode(0);
    gROOT->GetListOfCanvases()->Add(fCanvas);
    //fCanvas->SetBorderSize(1);
    //fCanvas->SetBit(kNoContextMenu);
    //fCanvas->SetFillColor(16);

    TGLayoutHints *lay = new TGLayoutHints(kLHintsExpandX);
    fUserFrame->AddFrame(vf1, lay);
    fUserFrame->AddFrame(vf2);
}

// --------------------------------------------------------------------------
//
//  Checks if the event number is valid, and if so reads the new event
//  and updates the display
//
void MEventDisplay::ReadinEvent(Int_t dir)
{
    MParList    *plist = (MParList*)   fEvtLoop->GetParList();
    MTaskList   *tlist = (MTaskList*)  fEvtLoop->GetTaskList();
    MGeomCam    *geom  = (MGeomCam*)   plist->FindObject("MGeomCam");
    MRawEvtData *raw   = (MRawEvtData*)plist->FindObject("MRawEvtData");

    //
    // Read first event.
    //
    MReadTree *reader = (MReadTree*)fEvtLoop->FindTask("MRead");

    const Int_t num = reader->GetNumEntry();

    do
    {
        if (dir<0 && !reader->DecEventNum())
        {
            reader->SetEventNum(num);
            return;
        }
        if (dir>0 && !reader->IncEventNum())
        {
            reader->SetEventNum(num);
            return;
        }

        const Int_t rc = tlist->Process();
        if (rc==kFALSE || rc==kERROR)
            return;

        reader->DecEventNum();

    } while (raw && raw->GetNumPixels()<1 && dir!=0);

    //
    // Cleare the 'FADC canvas'
    //
    fCanvas->Clear();
    fCanvas->Modified();
    fCanvas->Update();

    //
    // Print parameters
    //
    ((MHillas*)     plist->FindObject("MHillas"))->Print(*geom);
    ((MHillasExt*)  plist->FindObject("MHillasExt"))->Print(*geom);
    ((MHillasSrc*)  plist->FindObject("MHillasSrc"))->Print(*geom);
    plist->FindObject("MImagePar")->Print();
    ((MNewImagePar*)plist->FindObject("MNewImagePar"))->Print(*geom);
    plist->FindObject("MMuonCalibPar")->Print();
    ((MMuonSearchPar*)plist->FindObject("MMuonSearchPar"))->Print(*geom);

    //
    // UpdateDisplay
    //
    gStyle->SetOptStat(1101);
    Update();

    TGTextEntry *entry = (TGTextEntry*)fList->FindWidget(kEvtNumber);
    entry->SetText(Form("%d", reader->GetNumEntry()+1));
}

// --------------------------------------------------------------------------
//
//  Read first event to get display booted
//
void MEventDisplay::ReadFirstEvent()
{
    const Int_t rc = fEvtLoop->PreProcess();
    if (rc==kFALSE || rc==kERROR)
        return;

    UnLock();

    //
    // Get parlist
    //
    MParList *plist = (MParList*)fEvtLoop->GetParList();

    //
    // Add Geometry tab
    //
    AddGeometryTabs();

    //
    // Now read event...
    //
    ReadinEvent();

    MReadTree *reader = (MReadTree*)fEvtLoop->FindTask("MRead");
    TGString *txt = new TGString(Form("of %d", reader->GetEntries()));
    fNumOfEvts->SetText(txt);

    //
    // Draw ellipse on top of all pads
    //
    TObject *hillas1 = plist->FindObject("MHillas");
    TObject *hillas2 = plist->FindObject("MHillasSrc");
    TObject *hmuon   = plist->FindObject("MMuonSearchPar");
    for (int i=1; i<7;i++)
    {
        TCanvas *c = GetCanvas(i);
        c->GetPad(1)->cd(1);
        hmuon->Draw();
        hillas1->Draw();
        hillas2->Draw();
    }
}

// --------------------------------------------------------------------------
//
//  Adds the geometry tab 
//
void MEventDisplay::AddGeometryTabs()
{
    MGeomCam *geom = (MGeomCam*)fEvtLoop->GetParList()->FindObject("MGeomCam");
    if (!geom)
        return;

    TCanvas &c1=AddTab("Geometry");

    MHCamera *cam = new MHCamera(*geom);
    cam->SetBit(TH1::kNoStats|MHCamera::kNoLegend|kCanDelete);
    cam->Draw("pixelindex");

    c1.Modified();
    c1.Update();

    TCanvas &c2=AddTab("Sectors");

    cam = new MHCamera(*geom);
    cam->SetBit(TH1::kNoStats|MHCamera::kNoLegend|kCanDelete);
    cam->Draw("sectorindex");

    c2.Modified();
    c2.Update();
}

// --------------------------------------------------------------------------
//
//  ProcessMessage(Long_t msg, Long_t parm1, Long_t parm2)
//
// Processes information from all GUI items.
// Selecting an item usually generates an event with 4 parameters.
// The first two are packed into msg (first and second bytes).
// The other two are parm1 and parm2.
//
Bool_t MEventDisplay::ProcessMessage(Long_t msg, Long_t mp1, Long_t mp2)
{
    switch (GET_MSG(msg))
    {
    case kC_TEXTENTRY:
        switch(GET_SUBMSG(msg))
        {
        case kTE_ENTER:
            switch(GET_SUBMSG(msg))
            {
            case kTE_ENTER:
                {
                    TGTextEntry *entry = (TGTextEntry*)fList->FindWidget(kEvtNumber);
                    MReadTree *reader  = (MReadTree*)fEvtLoop->FindTask("MRead");
                    if (reader->SetEventNum(atoi(entry->GetText())-1))
                        ReadinEvent();
                }
                return kTRUE;
            }
            return kTRUE;
        }
        break;

    case kC_COMMAND:
        switch (GET_SUBMSG(msg))
        {
        case kCM_TAB:
            {
                //
                // Set name for 'FADC canvas'. The name is the anchor for MHCamera.
                // and clear the canvas
                TCanvas *c = GetCanvas(mp1);
                if (!c)
                    break;
                MHEvent *o = (MHEvent*)fEvtLoop->GetParList()->FindObject(c->GetName());
                if (!o)
                    break;
                fCanvas->SetName(Form("%p;%p;PixelContent", o->GetHist(), c->GetPad(1)->GetPad(1)));
            }
            break;

        case kCM_BUTTON:
            switch (mp1)
            {
            case kEvtPrev:
                ReadinEvent(-1);
                return kTRUE;

            case kEvtNext:
                ReadinEvent(+1);
                return kTRUE;
            }
            return kTRUE;
        }
        break;
    }

    return MStatusDisplay::ProcessMessage(msg, mp1, mp2);
}
